#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime, time

import pendulum
import pytest
import time_machine

from airflow.exceptions import TaskDeferred
from airflow.models.dag import DAG
from airflow.providers.standard.sensors.time import TimeSensor, TimeSensorAsync
from airflow.providers.standard.triggers.temporal import DateTimeTrigger
from airflow.utils import timezone

DEFAULT_TIMEZONE = "Asia/Singapore"  # UTC+08:00
DEFAULT_DATE_WO_TZ = datetime(2015, 1, 1)
DEFAULT_DATE_WITH_TZ = datetime(2015, 1, 1, tzinfo=timezone.parse_timezone(DEFAULT_TIMEZONE))


class TestTimeSensor:
    @pytest.mark.parametrize(
        "default_timezone, start_date, expected",
        [
            ("UTC", DEFAULT_DATE_WO_TZ, True),
            ("UTC", DEFAULT_DATE_WITH_TZ, False),
            (DEFAULT_TIMEZONE, DEFAULT_DATE_WO_TZ, False),
        ],
    )
    @time_machine.travel(timezone.datetime(2020, 1, 1, 23, 0).replace(tzinfo=timezone.utc))
    def test_timezone(self, default_timezone, start_date, expected, monkeypatch):
        monkeypatch.setattr("airflow.settings.TIMEZONE", timezone.parse_timezone(default_timezone))
        dag = DAG("test", schedule=None, default_args={"start_date": start_date})
        op = TimeSensor(task_id="test", target_time=time(10, 0), dag=dag)
        assert op.poke(None) == expected


class TestTimeSensorAsync:
    @time_machine.travel("2020-07-07 00:00:00", tick=False)
    def test_task_is_deferred(self):
        with DAG(
            dag_id="test_task_is_deferred",
            schedule=None,
            start_date=timezone.datetime(2020, 1, 1, 23, 0),
        ):
            op = TimeSensorAsync(task_id="test", target_time=time(10, 0))
        assert not timezone.is_naive(op.target_datetime)

        with pytest.raises(TaskDeferred) as exc_info:
            op.execute({})

        assert isinstance(exc_info.value.trigger, DateTimeTrigger)
        assert exc_info.value.trigger.moment == timezone.datetime(2020, 7, 7, 10)
        assert exc_info.value.kwargs is None
        assert exc_info.value.method_name == "execute_complete"

    def test_target_time_aware(self):
        with DAG("test_target_time_aware", schedule=None, start_date=timezone.datetime(2020, 1, 1, 23, 0)):
            aware_time = time(0, 1).replace(tzinfo=pendulum.local_timezone())
            op = TimeSensorAsync(task_id="test", target_time=aware_time)
            assert op.target_datetime.tzinfo == timezone.utc

    def test_target_time_naive_dag_timezone(self):
        """
        Tests that naive target_time gets converted correctly using the DAG's timezone.
        """
        with DAG(
            dag_id="test_target_time_naive_dag_timezone",
            schedule=None,
            start_date=pendulum.datetime(2020, 1, 1, 0, 0, tz=DEFAULT_TIMEZONE),
        ):
            op = TimeSensorAsync(task_id="test", target_time=pendulum.time(9, 0))
            assert op.target_datetime.time() == pendulum.time(1, 0)
            assert op.target_datetime.tzinfo == timezone.utc



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import inspect
from contextlib import closing
from datetime import datetime
from unittest import mock

import pytest

from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.models.dag import DAG
from airflow.utils import timezone

from tests_common.test_utils.compat import GenericTransfer
from tests_common.test_utils.providers import get_provider_min_airflow_version

pytestmark = pytest.mark.db_test

DEFAULT_DATE = timezone.datetime(2015, 1, 1)
DEFAULT_DATE_ISO = DEFAULT_DATE.isoformat()
DEFAULT_DATE_DS = DEFAULT_DATE_ISO[:10]
TEST_DAG_ID = "unit_test_dag"


@pytest.mark.backend("mysql")
class TestMySql:
    def setup_method(self):
        args = {"owner": "airflow", "start_date": DEFAULT_DATE}
        dag = DAG(TEST_DAG_ID, schedule=None, default_args=args)
        self.dag = dag

    def teardown_method(self):
        from airflow.providers.mysql.hooks.mysql import MySqlHook

        drop_tables = {"test_mysql_to_mysql", "test_airflow"}
        with closing(MySqlHook().get_conn()) as conn:
            for table in drop_tables:
                # Previous version tried to run execute directly on dbapi call, which was accidentally working
                with closing(conn.cursor()) as cur:
                    cur.execute(f"DROP TABLE IF EXISTS {table}")

    @pytest.mark.parametrize(
        "client",
        [
            "mysqlclient",
            "mysql-connector-python",
        ],
    )
    def test_mysql_to_mysql(self, client):
        class MySqlContext:
            def __init__(self, client):
                from airflow.providers.mysql.hooks.mysql import MySqlHook

                self.client = client
                self.connection = MySqlHook.get_connection(MySqlHook.default_conn_name)
                self.init_client = self.connection.extra_dejson.get("client", "mysqlclient")

            def __enter__(self):
                self.connection.set_extra(f'{{"client": "{self.client}"}}')

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.connection.set_extra(f'{{"client": "{self.init_client}"}}')

        with MySqlContext(client):
            sql = "SELECT * FROM connection;"
            op = GenericTransfer(
                task_id="test_m2m",
                preoperator=[
                    "DROP TABLE IF EXISTS test_mysql_to_mysql",
                    "CREATE TABLE IF NOT EXISTS test_mysql_to_mysql LIKE connection",
                ],
                source_conn_id="airflow_db",
                destination_conn_id="airflow_db",
                destination_table="test_mysql_to_mysql",
                sql=sql,
                dag=self.dag,
            )
            op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE, ignore_ti_state=True)

    @mock.patch("airflow.providers.common.sql.hooks.sql.DbApiHook.insert_rows")
    def test_mysql_to_mysql_replace(self, mock_insert):
        sql = "SELECT * FROM connection LIMIT 10;"
        op = GenericTransfer(
            task_id="test_m2m",
            preoperator=[
                "DROP TABLE IF EXISTS test_mysql_to_mysql",
                "CREATE TABLE IF NOT EXISTS test_mysql_to_mysql LIKE connection",
            ],
            source_conn_id="airflow_db",
            destination_conn_id="airflow_db",
            destination_table="test_mysql_to_mysql",
            sql=sql,
            dag=self.dag,
            insert_args={"replace": True},
        )
        op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE, ignore_ti_state=True)
        assert mock_insert.called
        _, kwargs = mock_insert.call_args
        assert "replace" in kwargs


@pytest.mark.backend("postgres")
class TestPostgres:
    def teardown_method(self):
        from airflow.providers.postgres.hooks.postgres import PostgresHook

        tables_to_drop = ["test_postgres_to_postgres", "test_airflow"]
        with PostgresHook().get_conn() as conn:
            with conn.cursor() as cur:
                for table in tables_to_drop:
                    cur.execute(f"DROP TABLE IF EXISTS {table}")

    def test_postgres_to_postgres(self, dag_maker):
        sql = "SELECT * FROM INFORMATION_SCHEMA.TABLES LIMIT 100;"
        with dag_maker(default_args={"owner": "airflow", "start_date": DEFAULT_DATE}, serialized=True):
            op = GenericTransfer(
                task_id="test_p2p",
                preoperator=[
                    "DROP TABLE IF EXISTS test_postgres_to_postgres",
                    "CREATE TABLE IF NOT EXISTS test_postgres_to_postgres (LIKE INFORMATION_SCHEMA.TABLES)",
                ],
                source_conn_id="postgres_default",
                destination_conn_id="postgres_default",
                destination_table="test_postgres_to_postgres",
                sql=sql,
            )
        dag_maker.create_dagrun()
        op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE, ignore_ti_state=True)

    @mock.patch("airflow.providers.common.sql.hooks.sql.DbApiHook.insert_rows")
    def test_postgres_to_postgres_replace(self, mock_insert, dag_maker):
        sql = "SELECT id, conn_id, conn_type FROM connection LIMIT 10;"
        with dag_maker(default_args={"owner": "airflow", "start_date": DEFAULT_DATE}, serialized=True):
            op = GenericTransfer(
                task_id="test_p2p",
                preoperator=[
                    "DROP TABLE IF EXISTS test_postgres_to_postgres",
                    "CREATE TABLE IF NOT EXISTS test_postgres_to_postgres (LIKE connection INCLUDING INDEXES)",
                ],
                source_conn_id="postgres_default",
                destination_conn_id="postgres_default",
                destination_table="test_postgres_to_postgres",
                sql=sql,
                insert_args={
                    "replace": True,
                    "target_fields": ("id", "conn_id", "conn_type"),
                    "replace_index": "id",
                },
            )
        dag_maker.create_dagrun()
        op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE, ignore_ti_state=True)
        assert mock_insert.called
        _, kwargs = mock_insert.call_args
        assert "replace" in kwargs


class TestGenericTransfer:
    def test_templated_fields(self):
        dag = DAG(
            "test_dag",
            schedule=None,
            start_date=datetime(2024, 10, 10),
            render_template_as_native_obj=True,
        )
        operator = GenericTransfer(
            task_id="test_task",
            sql="{{ sql }}",
            destination_table="{{ destination_table }}",
            source_conn_id="{{ source_conn_id }}",
            destination_conn_id="{{ destination_conn_id }}",
            preoperator="{{ preoperator }}",
            insert_args="{{ insert_args }}",
            dag=dag,
        )
        operator.render_template_fields(
            {
                "sql": "my_sql",
                "destination_table": "my_destination_table",
                "source_conn_id": "my_source_conn_id",
                "destination_conn_id": "my_destination_conn_id",
                "preoperator": "my_preoperator",
                "insert_args": {"commit_every": 5000, "executemany": True, "replace": True},
            }
        )
        assert operator.sql == "my_sql"
        assert operator.destination_table == "my_destination_table"
        assert operator.source_conn_id == "my_source_conn_id"
        assert operator.destination_conn_id == "my_destination_conn_id"
        assert operator.preoperator == "my_preoperator"
        assert operator.insert_args == {"commit_every": 5000, "executemany": True, "replace": True}

    def test_when_provider_min_airflow_version_is_3_0_or_higher_remove_obsolete_method(self):
        """
        Once this test starts failing due to the fact that the minimum Airflow version is now 3.0.0 or higher
        for this provider, you should remove the obsolete get_hook method in the GenericTransfer and use the
        one from BaseHook and remove this test.  This test was added to make sure to not forget to remove the
        fallback code for backward compatibility with Airflow 2.8.x which isn't need anymore once this
        provider depends on Airflow 3.0.0 or higher.
        """
        min_airflow_version = get_provider_min_airflow_version("apache-airflow-providers-standard")

        # Check if the current Airflow version is 3.0.0 or higher
        if min_airflow_version[0] >= 3:
            method_source = inspect.getsource(GenericTransfer.get_hook)
            raise AirflowProviderDeprecationWarning(
                f"Remove obsolete get_hook method in GenericTransfer:\n\r\n\r\t\t\t{method_source}"
            )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime

import pytest

from airflow.models.taskinstance import TaskInstance as TI
from airflow.providers.standard.operators.branch import BaseBranchOperator
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.timetables.base import DataInterval
from airflow.utils import timezone
from airflow.utils.state import State
from airflow.utils.task_group import TaskGroup
from airflow.utils.types import DagRunType

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.exceptions import DownstreamTasksSkipped
    from airflow.utils.types import DagRunTriggeredByType

pytestmark = pytest.mark.db_test

DEFAULT_DATE = timezone.datetime(2016, 1, 1)
INTERVAL = datetime.timedelta(hours=12)


class ChooseBranchOne(BaseBranchOperator):
    def choose_branch(self, context):
        return "branch_1"


class ChooseBranchOneTwo(BaseBranchOperator):
    def choose_branch(self, context):
        return ["branch_1", "branch_2"]


class ChooseBranchThree(BaseBranchOperator):
    def choose_branch(self, context):
        return ["branch_3"]


class TestBranchOperator:
    def test_without_dag_run(self, dag_maker):
        """This checks the defensive against non-existent tasks in a dag run"""
        dag_id = "branch_operator_test"
        triggered_by_kwargs = {"triggered_by": DagRunTriggeredByType.TEST} if AIRFLOW_V_3_0_PLUS else {}
        with dag_maker(
            dag_id,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")
            branch_op = ChooseBranchOne(task_id="make_choice")
            branch_1.set_upstream(branch_op)
            branch_2.set_upstream(branch_op)
        dag_maker.create_dagrun(**triggered_by_kwargs)

        if AIRFLOW_V_3_0_PLUS:
            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                branch_op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

            assert exc_info.value.tasks == [("branch_2", -1)]
        else:
            branch_op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

            for ti in dag_maker.session.query(TI).filter(
                TI.dag_id == dag_id, TI.execution_date == DEFAULT_DATE
            ):
                if ti.task_id == "make_choice":
                    assert ti.state == State.SUCCESS
                elif ti.task_id == "branch_1":
                    # should exist with state None
                    assert ti.state == State.NONE
                elif ti.task_id == "branch_2":
                    assert ti.state == State.SKIPPED
                else:
                    raise Exception

    def test_branch_list_without_dag_run(self, dag_maker):
        """This checks if the BranchOperator supports branching off to a list of tasks."""
        dag_id = "branch_operator_test"
        triggered_by_kwargs = {"triggered_by": DagRunTriggeredByType.TEST} if AIRFLOW_V_3_0_PLUS else {}
        with dag_maker(
            dag_id,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")
            branch_3 = EmptyOperator(task_id="branch_3")
            branch_op = ChooseBranchOneTwo(task_id="make_choice")
            branch_1.set_upstream(branch_op)
            branch_2.set_upstream(branch_op)
            branch_3.set_upstream(branch_op)
        dag_maker.create_dagrun(**triggered_by_kwargs)

        if AIRFLOW_V_3_0_PLUS:
            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                branch_op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

            assert exc_info.value.tasks == [("branch_3", -1)]
        else:
            branch_op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

            expected = {
                "make_choice": State.SUCCESS,
                "branch_1": State.NONE,
                "branch_2": State.NONE,
                "branch_3": State.SKIPPED,
            }

            for ti in dag_maker.session.query(TI).filter(
                TI.dag_id == dag_id, TI.execution_date == DEFAULT_DATE
            ):
                if ti.task_id in expected:
                    assert ti.state == expected[ti.task_id]
                else:
                    raise Exception

    def test_with_dag_run(self, dag_maker):
        dag_id = "branch_operator_test"
        triggered_by_kwargs = {"triggered_by": DagRunTriggeredByType.TEST} if AIRFLOW_V_3_0_PLUS else {}
        with dag_maker(
            dag_id,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")
            branch_op = ChooseBranchOne(task_id="make_choice")
            branch_1.set_upstream(branch_op)
            branch_2.set_upstream(branch_op)
        if AIRFLOW_V_3_0_PLUS:
            dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                logical_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )

            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                branch_op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

            assert exc_info.value.tasks == [("branch_2", -1)]
        else:
            dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                execution_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )

            branch_op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

            expected = {
                "make_choice": State.SUCCESS,
                "branch_1": State.NONE,
                "branch_2": State.SKIPPED,
            }

            for ti in dag_maker.session.query(TI).filter(
                TI.dag_id == dag_id, TI.execution_date == DEFAULT_DATE
            ):
                if ti.task_id in expected:
                    assert ti.state == expected[ti.task_id]
                else:
                    raise Exception

    def test_with_skip_in_branch_downstream_dependencies(self, dag_maker):
        dag_id = "branch_operator_test"
        triggered_by_kwargs = {"triggered_by": DagRunTriggeredByType.TEST} if AIRFLOW_V_3_0_PLUS else {}
        with dag_maker(
            dag_id,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")
            branch_op = ChooseBranchOne(task_id="make_choice")
            branch_op >> branch_1 >> branch_2
            branch_op >> branch_2

        if AIRFLOW_V_3_0_PLUS:
            dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                logical_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )
        else:
            dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                execution_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )

        branch_op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

        expected = {
            "make_choice": State.SUCCESS,
            "branch_1": State.NONE,
            "branch_2": State.NONE,
        }

        ti_date = TI.logical_date if AIRFLOW_V_3_0_PLUS else TI.execution_date

        for ti in dag_maker.session.query(TI).filter(TI.dag_id == dag_id, ti_date == DEFAULT_DATE):
            if ti.task_id in expected:
                assert ti.state == expected[ti.task_id]
            else:
                raise Exception

    def test_xcom_push(self, dag_maker):
        dag_id = "branch_operator_test"
        triggered_by_kwargs = {"triggered_by": DagRunTriggeredByType.TEST} if AIRFLOW_V_3_0_PLUS else {}
        with dag_maker(
            dag_id,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")
            branch_op = ChooseBranchOne(task_id="make_choice")
            branch_1.set_upstream(branch_op)
            branch_2.set_upstream(branch_op)

        if AIRFLOW_V_3_0_PLUS:
            dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                logical_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )

            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                branch_op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

            assert exc_info.value.tasks == [("branch_2", -1)]
        else:
            dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                execution_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )
            branch_op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

            ti_date = TI.logical_date if AIRFLOW_V_3_0_PLUS else TI.execution_date

            for ti in dag_maker.session.query(TI).filter(TI.dag_id == dag_id, ti_date == DEFAULT_DATE):
                if ti.task_id == "make_choice":
                    assert ti.xcom_pull(task_ids="make_choice") == "branch_1"

    def test_with_dag_run_task_groups(self, dag_maker):
        dag_id = "branch_operator_test"
        triggered_by_kwargs = {"triggered_by": DagRunTriggeredByType.TEST} if AIRFLOW_V_3_0_PLUS else {}
        with dag_maker(
            dag_id,
            default_args={"owner": "airflow", "start_date": DEFAULT_DATE},
            schedule=INTERVAL,
            serialized=True,
        ):
            branch_1 = EmptyOperator(task_id="branch_1")
            branch_2 = EmptyOperator(task_id="branch_2")
            branch_3 = TaskGroup("branch_3")
            EmptyOperator(task_id="task_1", task_group=branch_3)
            EmptyOperator(task_id="task_2", task_group=branch_3)
            branch_op = ChooseBranchThree(task_id="make_choice")
            branch_1.set_upstream(branch_op)
            branch_2.set_upstream(branch_op)
            branch_3.set_upstream(branch_op)

        if AIRFLOW_V_3_0_PLUS:
            dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                logical_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )

            with pytest.raises(DownstreamTasksSkipped) as exc_info:
                branch_op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

            assert set(exc_info.value.tasks) == {("branch_1", -1), ("branch_2", -1)}
        else:
            dag_maker.create_dagrun(
                run_type=DagRunType.MANUAL,
                start_date=timezone.utcnow(),
                execution_date=DEFAULT_DATE,
                state=State.RUNNING,
                data_interval=DataInterval(DEFAULT_DATE, DEFAULT_DATE),
                **triggered_by_kwargs,
            )
            branch_op.run(start_date=DEFAULT_DATE, end_date=DEFAULT_DATE)

            for ti in dag_maker.session.query(TI).filter(
                TI.dag_id == dag_id, TI.execution_date == DEFAULT_DATE
            ):
                if ti.task_id == "make_choice":
                    assert ti.state == State.SUCCESS
                elif ti.task_id == "branch_1":
                    assert ti.state == State.SKIPPED
                elif ti.task_id == "branch_2":
                    assert ti.state == State.SKIPPED
                elif ti.task_id == "branch_3.task_1":
                    assert ti.state == State.NONE
                elif ti.task_id == "branch_3.task_2":
                    assert ti.state == State.NONE
                else:
                    raise Exception
